from typing import Type

import numpy as np

from ..waveforms import Waveforms
from .cycles import CycleFeatureExtractor


def calculate_absolute_diffs(
    wf: Waveforms,
    name: str,
    feature_extractor: Type[CycleFeatureExtractor],
) -> Waveforms:
    """Calculates the absolute values of the differences between successive
    values of cycle-level features.

    waveforms.features.diffs[`name`][`feature_extractor.class_name`] is the same
    length as waveforms.features.cycles[`name`][`feature_extractor.class_name`]
    as always has first element 0.0

    Args:
        waveforms: `medical_waveforms.waveforms.Waveforms` instance holding your
            data
        name: Name of column in `waveforms` to extract feature from
        feature_extractor: Extractor class for the cycle-level feature you want
            to calculate the absolute differences for

    Returns:
        `waveforms` with the calculated absolute differences at
            waveforms.features.diffs[`name`][`feature_extractor.class_name`]
    """
    fe = feature_extractor()
    if not fe.class_name in wf.features.cycles[name].keys():
        wf = fe.extract_feature(wf, name)
    wf.features.diffs[name][fe.class_name] = np.zeros(
        wf.features.cycles[name][fe.class_name].size
    )
    wf.features.diffs[name][fe.class_name][1:] = np.abs(
        np.diff(wf.features.cycles[name][fe.class_name])
    )
    return wf



from abc import ABC, abstractmethod
from typing import List

import numpy as np
import pandas as pd

from ..waveforms import Waveforms


def get_cycles(waveforms: Waveforms, name: str) -> List[pd.DataFrame]:
    """Makes a list of the individual cycles from a waveform. This is useful
    for per-cycle feature extraction.

    Each cycle includes the troughs at its start and end.

    Args:
        waveforms: `medical_waveforms.waveforms.Waveforms` instance holding your data
        name: Name of column in `waveforms` to get cycles for

    Returns:
        Each element contains data from one cycle (e.g. a heartbeat)
    """
    return [
        waveforms.waveforms.iloc[
            waveforms.features.waveform[name]["troughs"][
                cycle_i
            ] : waveforms.features.waveform[name]["troughs"][cycle_i + 1]
            + 1
        ]
        for cycle_i in range(
            waveforms.features.waveform[name]["troughs"].size - 1
        )
    ]


class CycleFeatureExtractor(ABC):
    """Abstract base class for per-cycle (e.g. per heartbeat) feature
    extraction classes."""

    @property
    def class_name(self) -> str:
        """The name of the class itself."""
        return self.__class__.__name__

    @abstractmethod
    def extract_feature(self, waveforms: Waveforms, name: str) -> Waveforms:
        """Extracts a feature for each cycle in a waveform.

        Args:
            waveforms: `medical_waveforms.waveforms.Waveforms` instance holding your
                data
            name: Name of column in `waveforms` to extract feature from

        Returns:
            `waveforms` with new np.ndarray of shape (n_cycles,) at
                `waveforms.features.cycle[`name`][`self.class_name`]`
                where each element is the feature for the corresponding cycle
        """
        pass


class Duration(CycleFeatureExtractor):
    """Calculates duration (seconds) of each cycle in the waveform."""

    def extract_feature(self, waveforms: Waveforms, name: str) -> Waveforms:
        feature = [
            cycle[waveforms.time_column_name].values[-1]
            - cycle[waveforms.time_column_name].values[0]
            for cycle in get_cycles(waveforms, name)
        ]
        waveforms.features.cycles[name][self.class_name] = np.array(feature)
        return waveforms


class CyclesPerMinute(CycleFeatureExtractor):
    """Calculates rate (cycles per minute) for each cycle in the waveform."""

    def extract_feature(self, waveforms: Waveforms, name: str) -> Waveforms:
        feature = [
            cycle[waveforms.time_column_name].values[-1]
            - cycle[waveforms.time_column_name].values[0]
            for cycle in get_cycles(waveforms, name)
        ]
        waveforms.features.cycles[name][self.class_name] = 60 / np.array(
            feature
        )
        return waveforms


class MaximumValue(CycleFeatureExtractor):
    """Calculates maximum value of each cycle in the waveform."""

    def extract_feature(self, waveforms: Waveforms, name: str) -> Waveforms:
        feature = [cycle[name].max() for cycle in get_cycles(waveforms, name)]
        waveforms.features.cycles[name][self.class_name] = np.array(feature)
        return waveforms


class MinimumValue(CycleFeatureExtractor):
    """Calculates minimum value of each cycle in the waveform."""

    def extract_feature(self, waveforms: Waveforms, name: str) -> Waveforms:
        feature = [cycle[name].min() for cycle in get_cycles(waveforms, name)]
        waveforms.features.cycles[name][self.class_name] = np.array(feature)
        return waveforms


class MaximumMinusMinimumValue(CycleFeatureExtractor):
    """Calculates maximum minus minimum value of each cycle in the waveform."""

    def extract_feature(self, waveforms: Waveforms, name: str) -> Waveforms:
        feature = [
            cycle[name].max() - cycle[name].min()
            for cycle in get_cycles(waveforms, name)
        ]
        waveforms.features.cycles[name][self.class_name] = np.array(feature)
        return waveforms


class MeanValue(CycleFeatureExtractor):
    """Calculates mean value of each cycle in the waveform."""

    def extract_feature(self, waveforms: Waveforms, name: str) -> Waveforms:
        feature = [cycle[name].mean() for cycle in get_cycles(waveforms, name)]
        waveforms.features.cycles[name][self.class_name] = np.array(feature)
        return waveforms


class MeanNegativeFirstDifference(CycleFeatureExtractor):
    """Calculates the mean of only the negative first differences for each cycle
    in the waveform. This feature is similar to `mean_dyneg` from
    https://bit.ly/3AwtazE"""

    def extract_feature(self, waveforms: Waveforms, name: str) -> Waveforms:
        feature = [
            np.mean(
                self._remove_zeros(
                    np.clip(
                        np.diff(cycle[name].values, n=1, axis=0),
                        a_min=None,
                        a_max=0,
                    )
                )
            )
            for cycle in get_cycles(waveforms, name)
        ]
        # TODO: See https://github.com/UCL-Chimera/medical-waveforms/issues/16

        waveforms.features.cycles[name][self.class_name] = np.array(feature)
        return waveforms

    @staticmethod
    def _remove_zeros(array: np.ndarray) -> np.ndarray:
        zeros_removed = array[array < 0]
        if zeros_removed.size > 0:
            return zeros_removed
        # if no negative differences, indicates poor quality waveform
        return np.array([-np.inf])



import math
import os
from typing import Tuple

import numpy as np
import pandas as pd
from scipy import interpolate

from .utils import get_root_directory


def make_waveform_generator_from_file(filepath: str) -> interpolate.interp1d:
    """Make scipy interpolation function from saved example waveform file. This
    function can be used to generate resampled synthetic waveforms.

    Args:
        filepath: Path to .npy file containing a 1D numpy array of the example
            waveform

    Returns:
        Interpolation function with domain [0, 1]
    """
    waveform = np.load(filepath)
    time = np.linspace(start=0, stop=1, num=waveform.size)
    return interpolate.interp1d(
        time, waveform, kind="cubic", fill_value="extrapolate"
    )


def make_generator_timestamps_and_inputs(
    cycles_per_minute: float, n_cycles_target: float, hertz: float
) -> Tuple[np.ndarray, np.ndarray]:
    """Make timestamps and corresponding inputs for a synthetic waveform
        generator function.

    Args:
        cycles_per_minute: Rate (cycles per minute)
        n_cycles_target: Target number of cycles. Doesn't have to be an
            integer. The number of cycles actually generated will be slightly
            different from the target unless each cycle fits exactly within a
            whole number of samples.
        hertz: Sampling rate (hertz)

    Returns:
        Timestamps (seconds)
        Inputs. The units are standardised such that 0 is the start of the
            generated waveform and 1 is the end.
    """
    seconds_per_cycle = 60.0 / cycles_per_minute
    dt = 1 / hertz
    fractional_samples_per_cycle = seconds_per_cycle / dt

    # Add 1 as we have a sample at time zero
    n_samples = round(fractional_samples_per_cycle * n_cycles_target) + 1

    # Minus 1 to account for the zero sample
    n_seconds = (n_samples - 1) * dt
    n_cycles_actual = (n_samples - 1) / fractional_samples_per_cycle
    n_cycle_starts = math.ceil(n_cycles_actual)

    start_times = np.zeros(n_cycle_starts)
    start_times[1:] = np.cumsum(
        np.repeat(seconds_per_cycle, n_cycle_starts - 1)
    )

    timestamps = np.linspace(0, n_seconds, n_samples, endpoint=True)

    # `insertion_i` is the indices where `start_times` would be inserted into
    #  `timestamps` to maintain order. Each element in `offsets` is the number
    #  of seconds that elapse in a cycle before the first sample in that cycle.
    insertion_i = np.searchsorted(timestamps, start_times)
    offsets = timestamps[insertion_i] - start_times

    samples_per_cycle = np.zeros(n_cycle_starts)
    samples_per_cycle[:-1] = insertion_i[1:] - insertion_i[:-1]
    samples_per_cycle[-1] = n_samples - samples_per_cycle.sum()
    samples_per_cycle = samples_per_cycle.astype(int)

    def cycle_timestamps(cycle_i: int) -> np.ndarray:
        return (
            np.linspace(
                start=0,
                stop=(samples_per_cycle[cycle_i] - 1) / hertz,
                num=samples_per_cycle[cycle_i],
            )
            + offsets[cycle_i]
        )

    inputs = (
        np.concatenate(
            [cycle_timestamps(cycle_i) for cycle_i in range(n_cycle_starts)]
        )
        / seconds_per_cycle
    )

    return timestamps, inputs


def synthetic_arterial_pressure_data(
    systolic_pressure: float,
    diastolic_pressure: float,
    heart_rate: float,
    n_beats_target: float,
    hertz: float,
) -> pd.DataFrame:
    """Make synthetic arterial pressure data.

    Args:
        systolic_pressure: Maximum systolic arterial pressure (mmHg)
        diastolic_pressure: Minimum diastolic arterial pressure (mmHg)
        heart_rate: Heart rate (beats per minute)
        n_beats_target: Target number of heartbeats. Doesn't have to be an
            integer. The number of beats actually generated will be slightly
            different from the target unless each beat fits exactly within a
            whole number of samples.
        hertz: Sampling rate (hertz)

    Returns:
        Synthetic arterial pressure data with corresponding timestamps.
    """
    assert systolic_pressure > diastolic_pressure

    timestamps, inputs = make_generator_timestamps_and_inputs(
        cycles_per_minute=heart_rate,
        n_cycles_target=n_beats_target,
        hertz=hertz,
    )
    waveform_filepath = os.path.join(
        get_root_directory(),
        "medical_waveforms",
        "data",
        "example_arterial_pressure_waveform.npy",
    )
    generator = make_waveform_generator_from_file(waveform_filepath)
    waveform = generator(inputs)

    # Scale
    waveform -= waveform.min()
    waveform /= waveform.max()
    waveform *= systolic_pressure - diastolic_pressure
    waveform += diastolic_pressure

    return pd.DataFrame({"time": timestamps, "pressure": waveform})



from typing import Optional
from warnings import warn

from pyampd import ampd

from ..waveforms import Waveforms


def find_troughs(
    waveforms: Waveforms, name: str, scale: Optional[int] = None
) -> Waveforms:
    """Finds indices of troughs in a waveform.

    Args:
        waveforms: `medical_waveforms.waveforms.Waveforms` instance holding your data
        name: Name of column in `waveforms` to find troughs in
        scale : The maximum scale window size is (2 * scale + 1) during trough
            finding. Higher values require more memory. If None, sets `scale`
            to number_of_timesteps // 2 but often much lower values are
            adequate if your data contain many cycle. E.g.
            https://link.springer.com/chapter/10.1007/978-3-319-65798-1_39 uses
            `scale` values around one quarter of the cycle length for#
            intracranial pressure waveforms.

    Returns:
        `waveforms` with trough indices added to
            `waveforms.features.waveform[`name`]['troughs']`
    """
    waveforms.waveforms[name] *= -1  # invert signal, so troughs become peaks

    try:
        waveforms.features.waveform[name]["troughs"] = ampd.find_peaks(
            x=waveforms.waveforms[name], scale=scale
        )
    except MemoryError:
        warn("Ran out of memory. Try setting `scale` to a lower value.")

    waveforms.waveforms[name] *= -1  # un-invert signal
    return waveforms



from typing import Dict, Tuple

import numpy as np
import pandas as pd


class Waveforms:
    """Holds waveforms for downstream processing."""

    def __init__(
        self, waveforms: pd.DataFrame, time_column_name: str = "time"
    ):
        """
        Args:
            waveforms: Contains a timestamps column (in seconds) and one or more
                waveform columns (arbitrary units)
            time_column_name: The name of the timestamps column in `waveforms`
        """
        self.waveforms = waveforms
        self.time_column_name = time_column_name
        self._validate_arguments()
        self.names = self._init_names()
        self.features = FeaturesContainer(self.names)

    def _validate_arguments(self):
        assert isinstance(
            self.waveforms, pd.DataFrame
        ), "`waveforms` must be a pandas DataFrame"
        assert self.time_column_name in self.waveforms.columns, (
            "`waveforms` must contain a column called "
            f"'{self.time_column_name}'"
        )

    def _init_names(self) -> Tuple[str, ...]:
        """Makes a tuple of names of the waveform-containing columns in
        self.waveforms"""
        return tuple(
            name
            for name in self.waveforms.columns
            if name is not self.time_column_name
        )


class FeaturesContainer:
    """Holds features of the waveform data."""

    def __init__(self, waveform_names: Tuple[str, ...]):
        """
        Args:
            waveform_names: Names of each of the waveform columns
        """
        self.waveform = self._init_features_container(waveform_names)
        self.cycles = self._init_features_container(waveform_names)
        self.diffs = self._init_features_container(waveform_names)

    @staticmethod
    def _init_features_container(
        waveform_names,
    ) -> Dict[str, Dict[str, np.ndarray]]:
        """Makes a holder for features (the features themselves haven't been
        derived yet)"""
        return {name: {} for name in waveform_names}



from typing import Optional, Type

import numpy as np
import pandas as pd
from pydantic import BaseModel, Extra

from medical_waveforms.features import cycles, diffs
from medical_waveforms.waveforms import Waveforms


class CycleCheck(BaseModel):
    """A signal quality check to run on a feature of each cycle in a signal.

    Args:
        feature: The class that extracts the feature for each cycle
        min: The minimum acceptable value of the feature in order for the check
            to pass for that cycle
        max: The maximum acceptable value of the feature in order for the check
            to pass for that cycle
        units: The units of `min` and `max` (optional and just for
            documentation purposes)
        description: A description of the check (optional and just for
            documentation purposes)
    """

    feature: Type[cycles.CycleFeatureExtractor]
    min: float = -np.inf
    max: float = np.inf
    units: Optional[str] = None
    description: Optional[str] = None


class DiffCheck(BaseModel):
    """A signal quality check to run on the absolute differences between
        features of adjacent cycles in a signal.

    Args:
        feature: The class that extracts the absolute differences feature
        threshold: The minimum acceptable value of absolute difference in order
            for the check to pass for that cycle
        units: The units of `threshold` (optional and just for documentation
            purposes)
        description: A description of the check (optional and just for
            documentation purposes)
    """

    feature: Type[cycles.CycleFeatureExtractor]
    threshold: float
    units: Optional[str] = None
    description: Optional[str] = None


class ArterialPressureChecks(BaseModel):
    """Some preset checks for use with adult human arterial pressure signals."""

    diastolic_pressure: Optional[CycleCheck] = CycleCheck(
        feature=cycles.MinimumValue, min=20.0, max=200.0, units="mmHg"
    )
    systolic_pressure: Optional[CycleCheck] = CycleCheck(
        feature=cycles.MaximumValue, min=30.0, max=300.0, units="mmHg"
    )
    mean_pressure: Optional[CycleCheck] = CycleCheck(
        feature=cycles.MeanValue, min=30.0, max=200.0, units="mmHg"
    )
    heart_rate: Optional[CycleCheck] = CycleCheck(
        feature=cycles.CyclesPerMinute,
        min=20.0,
        max=200.0,
        units="beats per minute",
    )
    pulse_pressure: Optional[CycleCheck] = CycleCheck(
        feature=cycles.MaximumMinusMinimumValue,
        min=20.0,
        max=250.0,
        units="mmHg",
    )
    mean_dyneg: Optional[CycleCheck] = CycleCheck(
        feature=cycles.MeanNegativeFirstDifference,
        min=-3.0,
        max=0.0,
        units="mmHg",
    )
    diastolic_pressure_diff: Optional[DiffCheck] = DiffCheck(
        feature=cycles.MinimumValue, threshold=20.0, units="mmHg"
    )
    systolic_pressure_diff: Optional[DiffCheck] = DiffCheck(
        feature=cycles.MaximumValue, threshold=20.0, units="mmHg"
    )
    beat_time_diff: Optional[DiffCheck] = DiffCheck(
        feature=cycles.Duration, threshold=0.5, units="seconds"
    )

    class Config:
        """Allow users to define extra checks on instantiation."""

        extra = Extra.allow


def check_cycles(
    waveforms: Waveforms, name: str, checks: Type[BaseModel]
) -> pd.DataFrame:
    """Runs signal quality checks for each cycle in a signal.

    Args:
        waveforms: `medical_waveforms.waveforms.Waveforms` instance holding your data
        name: Name of column in `waveforms` to perform signal quality checks on
        checks: The checks to run. This should subclass pydantic's `BaseModel`
            and should have attributes which are instances of `CycleCheck`
            and/or `DiffCheck`, each of which defines a check.

    Returns:
        DataFrame with one row for each cycle in the signal. Has one Boolean
            column for each check, which is True if the check passed for that
            cycle, else False. Also has an 'all' column which is positive if
            all checks passed for that cycle.
    """
    checked = {}

    for check_name, check in vars(checks).items():
        if isinstance(check, CycleCheck):
            # Flag unphysiological cycles
            feature_extractor = check.feature()
            waveforms = check.feature().extract_feature(waveforms, name)
            checked[check_name] = (
                waveforms.features.cycles[name][feature_extractor.class_name]
                > check.min
            ) & (
                waveforms.features.cycles[name][feature_extractor.class_name]
                < check.max
            )

        elif isinstance(check, DiffCheck):
            # Flag unphysiological cycle-to-cycle changes
            feature_extractor = check.feature()
            waveforms = diffs.calculate_absolute_diffs(
                waveforms, name, check.feature
            )
            checked[check_name] = (
                waveforms.features.diffs[name][feature_extractor.class_name]
                < check.threshold
            )

    checked_df = pd.DataFrame(checked)
    checked_df["all"] = checked_df.all(axis=1)

    return checked_df



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['medical_waveforms', 'medical_waveforms.features']

package_data = \
{'': ['*'], 'medical_waveforms': ['data/*']}

install_requires = \
['matplotlib>=3.5.1,<4.0.0',
 'numpy>=1.22,<2.0',
 'pandas>=1.3.4,<2.0.0',
 'pyampd>=0.0.1,<0.0.2',
 'pydantic>=1.9.2,<2.0.0',
 'scipy>=1.7.3,<2.0.0']

setup_kwargs = {
    'name': 'medical-waveforms',
    'version': '0.1.1',
    'description': 'Preprocessing and analysis of physiological waveforms',
    'long_description': None,
    'author': 'Finneas Catling',
    'author_email': 'f.catling@imperial.ac.uk',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/UCL-Chimera/medical-waveforms',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
